/**
 * Copyright (C) 2008 Mathieu Carbou <mathieu.carbou@gmail.com>
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *         http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package com.mycila.xmltool;

import com.mycila.xmltool.util.ValidationResult;
import com.mycila.xmltool.util.XMLDocValidator;
import org.w3c.dom.Attr;
import org.w3c.dom.CDATASection;
import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.NamedNodeMap;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;
import org.w3c.dom.Text;
import org.xml.sax.InputSource;

import javax.xml.XMLConstants;
import javax.xml.namespace.NamespaceContext;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Result;
import javax.xml.transform.Source;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMResult;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;
import java.io.BufferedWriter;
import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.OutputStreamWriter;
import java.io.Reader;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.io.Writer;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.Set;
import java.util.TreeSet;

/**
 * @author Mathieu Carbou (mathieu.carbou@gmail.com)
 */
public final class XMLDoc implements XMLTag {

    private final XMLDocDefinition definition;
    private Element current;

    XMLDoc(XMLDocDefinition definition) {
        this.definition = definition;
        this.current = definition.getRoot();
    }

    public NamespaceContext getContext() {
        return definition;
    }

    public boolean hasAttribute(String name) {
        return current.hasAttribute(name);
    }

    public boolean hasAttribute(String name, String relativeXpath, Object... arguments) {
        Element old = current;
        try {
            return gotoTag(relativeXpath, arguments).hasAttribute(name);
        }
        finally {
            current = old;
        }
    }

    public boolean hasTag(String relativeXpath, Object... arguments) {
        Element old = current;
        try {
            gotoTag(relativeXpath, arguments);
            return true;
        }
        catch (Exception e) {
            return false;
        }
        finally {
            current = old;
        }
    }

    public XMLTag forEachChild(CallBack callBack) {
        Element old = current;
        try {
            for (Element node : childs(current)) {
                current = node;
                callBack.execute(this);
            }
            return this;
        }
        finally {
            current = old;
        }
    }

    public XMLTag forEach(CallBack callBack, String relativeXpath, Object... arguments) {
        Element old = current;
        try {
            Node[] nodes = definition.getXpath().findNodes(current, relativeXpath, arguments);
            List<Element> els = new ArrayList<Element>(nodes.length);
            for (Node node : nodes) {
                if (isElement(node)) {
                    els.add((Element) node);
                }
            }
            for (Element node : els) {
                current = node;
                callBack.execute(this);
            }
            return this;
        }
        finally {
            current = old;
        }
    }

    //TODO: 3.0 - Remove deprecated method
    @Deprecated
    public XMLTag forEach(com.google.code.xmltool.CallBack callBack, String relativeXpath, Object... arguments) {
        return forEach(wrap(callBack), relativeXpath, arguments);
    }

    //TODO: 3.0 - Remove deprecated method
    @Deprecated
    public XMLTag forEachChild(com.google.code.xmltool.CallBack callBack) {
        return forEachChild(wrap(callBack));
    }

    //TODO: 3.0 - Remove deprecated method
    @Deprecated
    private CallBack wrap(final com.google.code.xmltool.CallBack cb) {
        return new CallBack() {
            public void execute(XMLTag doc) {
                cb.execute(doc);
            }
        };
    }

    public String rawXpathString(String relativeXpath, Object... arguments) {
        return definition.getXpath().rawXpathString(current, relativeXpath, arguments);
    }

    public Number rawXpathNumber(String relativeXpath, Object... arguments) {
        return definition.getXpath().rawXpathNumber(current, relativeXpath, arguments);
    }

    public Boolean rawXpathBoolean(String relativeXpath, Object... arguments) {
        return definition.getXpath().rawXpathBoolean(current, relativeXpath, arguments);
    }

    public Node rawXpathNode(String relativeXpath, Object... arguments) {
        return definition.getXpath().rawXpathNode(current, relativeXpath, arguments);
    }

    public NodeList rawXpathNodeSet(String relativeXpath, Object... arguments) {
        return definition.getXpath().rawXpathNodeSet(current, relativeXpath, arguments);
    }

    public String getPefix(String namespaceURI) {
        String prefix = getContext().getPrefix(namespaceURI);
        return prefix == null ? XMLConstants.DEFAULT_NS_PREFIX : prefix;
    }

    @SuppressWarnings({"unchecked"})
    public String[] getPefixes(String namespaceURI) {
        Set<String> prefixes = new TreeSet<String>();
        Iterator<String> i = getContext().getPrefixes(namespaceURI);
        while (i.hasNext()) {
            prefixes.add(i.next());
        }
        return prefixes.toArray(new String[prefixes.size()]);
    }

    public XMLTag addNamespace(String prefix, String namespaceURI) {
        definition.addNamespace(prefix, namespaceURI);
        return this;
    }

    public XMLTag addDocument(XMLTag tag) {
        return addDocument(tag.toDocument());
    }

    public XMLTag addDocument(Document doc) {
        current.appendChild(current.getOwnerDocument().importNode(doc.getDocumentElement(), true));
        return this;
    }

    public XMLTag addTag(XMLTag tag) {
        return addTag(tag.getCurrentTag());
    }

    public XMLTag addTag(Element tag) {
        current.appendChild(current.getOwnerDocument().importNode(tag, true));
        return this;
    }

    public XMLTag addTag(String name) {
        checkEmpty("Tag name", name);
        Element el = definition.createElement(name);
        current.appendChild(el);
        current = el;
        return this;
    }

    private void checkEmpty(String name, String value) {
        if (value == null || "".equals(value)) {
            throw new IllegalArgumentException(name + " cannot be empty");
        }
    }

    public XMLTag addAttribute(String name, String value) {
        checkEmpty("Attribute name", name);
        if (hasAttribute(name)) {
            throw new XMLDocumentException("Attribute '%s' already exist on tag '%s'", name, getCurrentTagName());
        }
        definition.createAttribute(current, name, value);
        return this;
    }

    public XMLTag addAttribute(Attr attr) {
        if (hasAttribute(attr.getName())) {
            throw new XMLDocumentException("Attribute '%s' already exist on tag '%s'", attr.getName(), getCurrentTagName());
        }
        current.setAttributeNodeNS((Attr) current.getOwnerDocument().importNode(attr, true));
        return this;
    }

    public XMLTag addText(String text) {
        current.appendChild(definition.createText(text));
        return gotoParent();
    }

    public XMLTag addText(Text text) {
        current.appendChild(current.getOwnerDocument().importNode(text, true));
        return gotoParent();
    }

    public XMLTag addCDATA(String data) {
        current.appendChild(definition.createCDATA(data));
        return gotoParent();
    }

    public XMLTag addCDATA(CDATASection data) {
        current.appendChild(current.getOwnerDocument().importNode(data, true));
        return gotoParent();
    }

    public XMLTag delete() {
        if (current == definition.getRoot()) {
            throw new XMLDocumentException("Cannot delete root node '%s'", getCurrentTagName());
        }
        Node toDelete = current;
        gotoParent();
        current.removeChild(toDelete);
        return this;
    }

    public XMLTag deleteChilds() {
        Element[] toDelete = childs(current);
        for (Element node : toDelete) {
            current.removeChild(node);
        }
        return this;
    }

    public XMLTag deleteAttributes() {
        Attr[] attrs = attr(current);
        for (Attr attr : attrs) {
            current.removeAttributeNode(attr);
        }
        return this;
    }

    public XMLTag deleteAttribute(String name) {
        if (!hasAttribute(name)) {
            throw new XMLDocumentException("Cannot delete attribute '%s' from element '%s': attribute does noe exist", name, getCurrentTagName());
        }
        current.removeAttribute(name);
        return this;
    }

    public XMLTag renameTo(String newNodeName) {
        current = definition.rename(current, newNodeName);
        return this;
    }

    public XMLTag getInnerDocument() {
        return fromCurrentTag(this, definition.isIgnoreNamespaces());
    }

    public String getInnerText() {
        try {
            TransformerFactory tf = TransformerFactory.newInstance();
            /*try {
                tf.setAttribute("indent-number", 0);
            } catch (Exception ignored) {
            }*/
            Transformer transformer = tf.newTransformer();
            //transformer.setParameter("indent-number", 0);
            //transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
            //transformer.setOutputProperty(OutputKeys.METHOD, "xml");
            transformer.setOutputProperty(OutputKeys.ENCODING, definition.getEncoding());
            //transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            //transformer.setOutputProperty(OutputKeys.STANDALONE, definition.getDocument().getXmlStandalone() ? "yes" : "no");
            transformer.setOutputProperty(OutputKeys.OMIT_XML_DECLARATION, "yes");

            StringWriter sw = new StringWriter();
            NodeList list = current.getChildNodes();
            for (int i = 0; i < list.getLength(); i++) {
                transformer.transform(new DOMSource(list.item(i)), new StreamResult(sw));
            }
            
            return sw.toString();
        }
        catch (TransformerException e) {
            throw new XMLDocumentException("Transformation error", e);
        }

    }

    public XMLTag gotoParent() {
        Node n = current.getParentNode();
        if (current != definition.getRoot() && isElement(n)) {
            current = (Element) n;
        }
        return this;
    }

    public XMLTag gotoRoot() {
        current = definition.getRoot();
        return this;
    }

    public XMLTag gotoChild() {
        Element[] els = childs(current);
        switch (els.length) {
            case 0:
                throw new XMLDocumentException("Current element '%s' has no child", getCurrentTagName());
            case 1:
                current = els[0];
                break;
            default:
                throw new XMLDocumentException("Cannot select child: current element '%s' has '%s' children", getCurrentTagName(), els.length);
        }
        return this;
    }

    public XMLTag gotoChild(int i) {
        Element[] els = childs(current);
        if (i > 0 && i <= els.length) {
            current = els[i - 1];
        } else {
            throw new XMLDocumentException("Cannot acces child '%s' of element '%s' amongst its '%s' childs", i, getCurrentTagName(), els.length);
        }
        return this;
    }

    public XMLTag gotoChild(String nodeName) {
        Element[] els = childs(current);
        List<Element> found = new ArrayList<Element>(els.length);
        for (Element el : els) {
            if (el.getTagName().equals(nodeName)) {
                found.add(el);
            }
        }
        switch (found.size()) {
            case 0:
                throw new XMLDocumentException("Current element '%s' has no child named '%s'", getCurrentTagName(), nodeName);
            case 1:
                current = found.get(0);
                break;
            default:
                throw new XMLDocumentException("Cannot select child: current element '%s' has '%s' children named '%s'", getCurrentTagName(), found.size(), nodeName);
        }
        return this;
    }

    public XMLTag gotoTag(String relativeXpath, Object... arguments) {
        Node n = definition.getXpath().findNode(current, relativeXpath, arguments);
        if (!isElement(n)) {
            throw new XMLDocumentException("XPath expression '%s' does not target an element. Targeted node is '%s' (node type is '%s')", String.format(relativeXpath, arguments), n.getNodeName(), n.getNodeType());
        }
        current = (Element) n;
        return this;
    }

    public Element getCurrentTag() {
        return current;
    }

    public String getCurrentTagName() {
        return current.getTagName();
    }

    public String getAttribute(String name) {
        if (!hasAttribute(name)) {
            throw new XMLDocumentException("Element '%s' does not have attribute '%s'", getCurrentTagName(), name);
        }
        return current.getAttribute(name);
    }

    public String getAttribute(String name, String relativeXpath, Object... arguments) {
        Element old = current;
        try {
            return gotoTag(relativeXpath, arguments).getAttribute(name);
        }
        finally {
            current = old;
        }
    }

    @SuppressWarnings({"MismatchedReadAndWriteOfArray"})
    public String[] getAttributeNames() {
        Attr[] attrs = attr(current);
        String[] names = new String[attrs.length];
        for (int i = 0; i < attrs.length; i++) {
            names[i] = (attrs[i]).getName();
        }
        return names;
    }

    public String getText(String relativeXpath, Object... arguments) {
        Element old = current;
        try {
            return gotoTag(relativeXpath, arguments).getText();
        }
        finally {
            current = old;
        }
    }

    public String getText() {
        List<Node> nodes = getChildNodes(Document.TEXT_NODE);
        if (nodes.isEmpty()) {
            return "";
        }
        StringBuilder sb = new StringBuilder();
        for (Node node : nodes) {
            String val = node.getNodeValue();
            if (val != null) {
                sb.append(val);
            }
        }
        return sb.toString();
    }

    public String getTextOrCDATA() {
        String txt = getText();
        return "".equals(txt) ? getCDATA() : txt;

    }

    public String getTextOrCDATA(String relativeXpath, Object... arguments) throws XMLDocumentException {
        String txt = getText(relativeXpath, arguments);
        return "".equals(txt) ? getCDATA(relativeXpath, arguments) : txt;

    }

    public String getCDATA(String relativeXpath, Object... arguments) {
        Element old = current;
        try {
            return gotoTag(relativeXpath, arguments).getCDATA();
        }
        finally {
            current = old;
        }
    }

    public String getCDATA() {
        StringBuilder sb = new StringBuilder();
        List<Node> nodes = getChildNodes(Document.CDATA_SECTION_NODE);
        for (Node node : nodes) {
            if (node.getNodeType() == Document.CDATA_SECTION_NODE) {
                String val = ((CDATASection) node).getData();
                if (val != null) {
                    sb.append(val);
                }
            }
        }
        return sb.toString();
    }

    public Document toDocument() {
        return definition.getDocument();
    }

    public Source toSource() {
        return new DOMSource(toDocument());
    }

    @Override
    public String toString() {
        return toString(definition.getEncoding());
    }

    public String toString(String encoding) {
        StringWriter out = new StringWriter();
        toStream(out, encoding);
        return out.toString();
    }

    public byte[] toBytes() {
        return toBytes(definition.getEncoding());
    }

    public byte[] toBytes(String encoding) {
        ByteArrayOutputStream out = new ByteArrayOutputStream();
        toStream(out, encoding);
        return out.toByteArray();
    }

    public XMLTag toStream(OutputStream out) {
        return toStream(out, definition.getEncoding());
    }

    public XMLTag toStream(OutputStream out, String encoding) {
        try {
            return toStream(new BufferedWriter(new OutputStreamWriter(out, encoding)));
        } catch (UnsupportedEncodingException e) {
            throw new XMLDocumentException(e.getMessage(), e);
        }
    }

    public XMLTag toStream(Writer out) {
        return toStream(out, definition.getEncoding());
    }

    public XMLTag toStream(Writer out, String encoding) {
        return toResult(new StreamResult(out), encoding);
    }

    public Result toResult() {
        Result r = new DOMResult();
        toResult(r);
        return r;
    }

    public Result toResult(String encoding) {
        Result r = new DOMResult();
        toResult(r, encoding);
        return r;
    }

    public OutputStream toOutputStream() {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        toStream(baos);
        return baos;
    }

    public OutputStream toOutputStream(String encoding) {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        toStream(baos, encoding);
        return baos;
    }

    public Writer toWriter() {
        Writer w = new StringWriter();
        toStream(w);
        return w;
    }

    public Writer toWriter(String encoding) {
        Writer w = new StringWriter();
        toStream(w, encoding);
        return w;
    }

    public XMLTag toResult(Result out) {
        return toResult(out, definition.getEncoding());
    }

    public XMLTag toResult(Result out, String encoding) {
        try {
            TransformerFactory tf = TransformerFactory.newInstance();
            try {
                tf.setAttribute("indent-number", 4);
            } catch (Exception ignored) {
            }
            Transformer transformer = tf.newTransformer();
            transformer.setParameter("indent-number", 4);
            transformer.setOutputProperty(OutputKeys.METHOD, "xml");
            transformer.setOutputProperty(OutputKeys.ENCODING, encoding);
            transformer.setOutputProperty(OutputKeys.INDENT, "yes");
            transformer.setOutputProperty(OutputKeys.STANDALONE, definition.getDocument().getXmlStandalone() ? "yes" : "no");
            transformer.setOutputProperty("{http://xml.apache.org/xslt}indent-amount", "4");
            transformer.transform(toSource(), out);
            return this;
        }
        catch (TransformerException e) {
            throw new XMLDocumentException("Transformation error", e);
        }
    }

    public ValidationResult validate(Source... schemas) {
        try {
            return XMLDocValidator.validate(this.toDocument(), schemas);
        }
        catch (Exception e) {
            throw new XMLDocumentException("Validation failed", e);
        }
    }

    public ValidationResult validate(URL... schemaLocations) {
        try {
            return XMLDocValidator.validate(this.toDocument(), schemaLocations);
        }
        catch (Exception e) {
            throw new XMLDocumentException("Validation failed", e);
        }
    }

    private boolean isElement(Node n) {
        return n != null && n.getNodeType() == Document.ELEMENT_NODE;
    }

    private Element[] childs(Element e) {
        NodeList list = e.getChildNodes();
        List<Element> els = new ArrayList<Element>(list.getLength());
        for (int i = 0; i < list.getLength(); i++) {
            if (isElement(list.item(i))) {
                els.add((Element) list.item(i));
            }
        }
        return els.toArray(new Element[els.size()]);
    }

    private Attr[] attr(Element e) {
        NamedNodeMap list = e.getAttributes();
        Attr[] attrs = new Attr[list.getLength()];
        for (int i = 0; i < list.getLength(); i++) {
            attrs[i] = (Attr) list.item(i);
        }
        return attrs;
    }

    private List<Node> getChildNodes(short type) {
        NodeList childs = current.getChildNodes();
        List<Node> nodes = new ArrayList<Node>(childs.getLength());
        for (int i = 0; i < childs.getLength(); i++) {
            if (childs.item(i).getNodeType() == type) {
                nodes.add(childs.item(i));
            }
        }
        return nodes;
    }

    public XMLTag duplicate() {
        XMLTag clone = XMLDoc.from(this, definition.isIgnoreNamespaces());
        StringBuilder sb = new StringBuilder();
        Element thisNode = current;
        while (thisNode != definition.getRoot()) {
            Element parent = (Element) thisNode.getParentNode();
            Element[] els = childs(parent);
            for (int i = 0; i < els.length; i++) {
                if (els[i] == thisNode) {
                    sb.insert(0, "/*[" + (i + 1) + "]");
                    break;
                }
            }
            thisNode = parent;
        }
        return clone.gotoRoot().gotoTag(sb.deleteCharAt(0).toString());
    }

    //TODO: 3.0 - Remove deprecated methods
    @Deprecated
    public XMLTag clone() {
        return duplicate();
    }

    public static XMLDocBuilder newDocument(boolean ignoreNamespaces) {
        return XMLDocBuilder.newDocument(ignoreNamespaces);
    }

    public static XMLTag from(Node node, boolean ignoreNamespaces) {
        return XMLDocBuilder.from(node, ignoreNamespaces);
    }

    public static XMLTag from(InputSource source, boolean ignoreNamespaces) {
        return XMLDocBuilder.from(source, ignoreNamespaces);
    }

    public static XMLTag from(Reader reader, boolean ignoreNamespaces) {
        return XMLDocBuilder.from(reader, ignoreNamespaces);
    }

    public static XMLTag from(InputStream is, boolean ignoreNamespaces) {
        return XMLDocBuilder.from(is, ignoreNamespaces);
    }

    public static XMLTag from(File file, boolean ignoreNamespaces) {
        return XMLDocBuilder.from(file, ignoreNamespaces);
    }

    public static XMLTag from(URL xmlLocation, boolean ignoreNamespaces) {
        return XMLDocBuilder.from(xmlLocation, ignoreNamespaces);
    }

    public static XMLTag from(String xmlData, boolean ignoreNamespaces) {
        return XMLDocBuilder.from(xmlData, ignoreNamespaces);
    }

    public static XMLTag from(Source source, boolean ignoreNamespaces) {
        return XMLDocBuilder.from(source, ignoreNamespaces);
    }

    public static XMLTag from(XMLTag tag, boolean ignoreNamespaces) {
        return XMLDocBuilder.from(tag, ignoreNamespaces);
    }

    public static XMLTag fromCurrentTag(XMLTag tag, boolean ignoreNamespaces) {
        return XMLDocBuilder.fromInner(tag, ignoreNamespaces);
    }
}
